#include "window.hh"
#include "window_events.hh"
#include "idlemgr.hh"

#include <typeinfo>

void window::redraw_after_move(const rect &oldr, const rect &newr)
{
	rect a, b; /* what's left over */
	oldr.clipboxes(newr, a, b);
	if(owner)
	{
		owner->redraw(a);    /* draw ghost piece 1 */
		owner->redraw(b);    /* draw ghost piece 2 */
		owner->redraw(newr); /* draw the new area (this window) */
	}
}
window::childlist::iterator window::find_child(const window *p)
{
	childlist::iterator i;
	for(i=children.begin(); i!=children.end(); ++i)if(i->p == p)break;
	return i;
}
window::childlist::const_iterator window::find_child(const window *p) const
{
	childlist::const_iterator i;
	for(i=children.begin(); i!=children.end(); ++i)if(i->p == p)break;
	return i;
}

void window::drawchildren(unsigned x1,unsigned y1, unsigned x2,unsigned y2)
{
	childlist::const_iterator i;
	for(i=children.begin(); i!=children.end(); ++i)
	{
		rect childwin(i->x,i->y, i->x+i->p->width()-1, i->y+i->p->height()-1);
		rect r = rect(x1,y1, x2,y2).intersect(childwin);
		if(r)
		{
			/*
			printf("\r\nHello, I am (%u,%u,%u,%u) at %p\r\n",
				getxpos(),getypos(), getxpos()+width()-1,getypos()+height()-1, this);
			printf("Intersection between:\r\n");
			printf("    (%u,%u,%u,%u)\r\n", x1,y1, x2,y2);
			printf("and (%u,%u,%u,%u) (child=%p)\r\n", childwin.x1,childwin.y1, childwin.x2,childwin.y2, &*i->p);
			printf("is: (%u,%u,%u,%u)\r\n", r.x1,r.y1, r.x2,r.y2);
			*/
			r.move(-i->x, -i->y);
			/*
			printf("Child gonna redraw (%u,%u,%u,%u)\r\n", r.x1,r.y1, r.x2,r.y2);
			*/
			i->p->redraw(r.x1, r.y1, r.x2, r.y2);
			
			buf.drawfrom(i->p->buf.piece(r.x1, r.y1, r.x2, r.y2), i->x, i->y);
		}
	}
}

void window::redraw(unsigned x1,unsigned y1, unsigned x2,unsigned y2)
{
	drawchildren(x1,y1, x2,y2);
}	

void window::resize(unsigned nxs, unsigned nys)
{
	unsigned oldwidth = width();
	unsigned oldheight = height();
	buf.resize(nxs, nys);
	unsigned x = getxpos(), y = getypos();
	rect oldr(x, y, x+oldwidth-1, y+oldheight-1); /* old area */
	rect newr(x, y, y+width()-1,  y+height()-1);  /* new area */
	redraw_after_move(oldr, newr);
}

void window::move(unsigned x, unsigned y) /* parameters: new position within owner */
{
	rect oldr(getxpos(), getypos(), getxpos() + width() - 1, getypos() + height() - 1);
	owner->movechild(this, x, y);
	rect newr(x, y, x + width() - 1, y + height() - 1);
	redraw_after_move(oldr, newr);
}
void window::addchild(winptr p, unsigned x, unsigned y)
{
	event e;
	
	p->owner = this;
	
	if(children.size())
	{
		e = new event_lostfocus;
		children.rbegin()->p->handleevent(e);
	}
	
	children.push_back(childwin(p, x, y));
	
	e = new event_gotfocus;
	/*children.rbegin()->*/p->handleevent(e);
}

/* Note that type event shows as an autoptr */
void window::handleevent(event &e)
{
	if(typeid(*e) == typeid(event_idle))return;
	
	if(typeid(*e) == typeid(event_gotfocus))
	{
		focused = true;
		e.clear();
		return;
	}
	if(typeid(*e) == typeid(event_lostfocus))
	{
		focused = false;
		e.clear();
		return;
	}
	if(!focused)return;

	childlist::reverse_iterator i;
	for(i = children.rbegin(); i != children.rend(); ++i)
	{
		i->p->handleevent(e);
		if(typeid(*e) == typeid(event_none))return;
	}	
	if(event_resize *p = dynamic_cast<event_resize *> (&*e))
	{
		int newx = width() + p->dx;
		int newy = height() + p->dy;
		if(newx<1)newx=1;
		if(newy<1)newy=1;
		resize(newx, newy);
		e.clear();
	}
	else if(event_move *p = dynamic_cast<event_move *> (&*e))
	{
		int newx = getxpos() + p->dx;
		int newy = getypos() + p->dy;
		if(newx<0)newx=0;
		if(newy<0)newy=0;
		move(newx, newy);
		e.clear();
	}
}
